@@ -6,11 +6,20 @@
  */
 
 #include "candidatelist.h"
-#include <functional>
+#include <algorithm>
+#include <cstddef>
+#include <memory>
 #include <stdexcept>
+#include <string>
 #include <unordered_set>
+#include <utility>
+#include <vector>
 #include <fcitx-utils/macros.h>
 #include <fcitx-utils/utf8.h>
+#include "fcitx-utils/key.h"
+#include "fcitx-utils/keysym.h"
+#include "fcitx-utils/misc.h"
+#include "text.h"
 
 namespace fcitx {
 
@@ -77,6 +86,8 @@ public:
 
 } // namespace
 
+ActionableCandidateList::~ActionableCandidateList() = default;
+
 class CandidateListPrivate {
 public:
     BulkCandidateList *bulk_ = nullptr;
@@ -85,6 +96,7 @@ public:
     CursorMovableCandidateList *cursorMovable_ = nullptr;
     BulkCursorCandidateList *bulkCursor_ = nullptr;
     CursorModifiableCandidateList *cursorModifiable_ = nullptr;
+    ActionableCandidateList *actionable_ = nullptr;
 };
 
 CandidateList::CandidateList()
@@ -124,6 +136,11 @@ BulkCursorCandidateList *CandidateList::toBulkCursor() const {
     return d->bulkCursor_;
 }
 
+ActionableCandidateList *CandidateList::toActionable() const {
+    FCITX_D();
+    return d->actionable_;
+}
+
 void CandidateList::setBulk(BulkCandidateList *list) {
     FCITX_D();
     d->bulk_ = list;
@@ -154,6 +171,11 @@ void CandidateList::setBulkCursor(BulkCursorCandidateList *list) {
     d->bulkCursor_ = list;
 }
 
+void CandidateList::setActionable(ActionableCandidateList *list) {
+    FCITX_D();
+    d->actionable_ = list;
+}
+
 class CandidateWordPrivate {
 public:
     CandidateWordPrivate(Text &&text) : text_(std::move(text)) {}
@@ -330,6 +352,7 @@ public:
     bool cursorKeepInSamePage_ = false;
     CursorPositionAfterPaging cursorPositionAfterPaging_ =
         CursorPositionAfterPaging::DonotChange;
+    std::unique_ptr<ActionableCandidateList> actionable_;
 
     int size() const {
         auto start = currentPage_ * pageSize_;
@@ -340,7 +363,9 @@ public:
         return remain;
     }
 
-    int toGlobalIndex(int idx) const { return idx + currentPage_ * pageSize_; }
+    int toGlobalIndex(int idx) const {
+        return idx + (currentPage_ * pageSize_);
+    }
 
     void checkIndex(int idx) const {
         if (idx < 0 || idx >= size()) {
@@ -402,7 +427,7 @@ std::string keyToLabel(const Key &key) {
 
 #define _APPEND_MODIFIER_STRING(STR, VALUE)                                    \
     if (key.states() & KeyState::VALUE) {                                      \
-        result += STR;                                                         \
+        result += (STR);                                                       \
     }
     _APPEND_MODIFIER_STRING("C-", Ctrl)
     _APPEND_MODIFIER_STRING("A-", Alt)
@@ -417,17 +442,17 @@ std::string keyToLabel(const Key &key) {
     } else {
         result = Key::keySymToString(key.sym(), KeyStringFormat::Localized);
     }
-    // add a dot as separator
-    result += ". ";
+    if (!isApple()) {
+        // add a dot as separator
+        result += ". ";
+    }
 
     return result;
 }
 
 void CommonCandidateList::setLabels(const std::vector<std::string> &labels) {
     FCITX_D();
-    fillLabels(
-        d->labels_, labels,
-        [](const std::string &str) -> const std::string & { return str; });
+    fillLabels(d->labels_, labels, [](const std::string &str) { return str; });
 }
 
 void CommonCandidateList::setSelectionKey(const KeyList &keyList) {
@@ -652,9 +677,10 @@ void CommonCandidateList::moveCursor(bool prev) {
                 setPage(d->cursorIndex_ / d->pageSize_);
             }
         }
-    } while (!deadloopDetect.count(d->cursorIndex_) && d->cursorIndex_ >= 0 &&
+    } while (!deadloopDetect.contains(d->cursorIndex_) &&
+             d->cursorIndex_ >= 0 &&
              candidateFromAll(d->cursorIndex_).isPlaceHolder());
-    if (deadloopDetect.count(d->cursorIndex_)) {
+    if (deadloopDetect.contains(d->cursorIndex_)) {
         d->cursorIndex_ = startCursor;
         d->currentPage_ = startPage;
     }
@@ -711,4 +737,12 @@ void CommonCandidateList::fixAfterUpdate() {
         }
     }
 }
+
+void CommonCandidateList::setActionableImpl(
+    std::unique_ptr<ActionableCandidateList> actionable) {
+    FCITX_D();
+    d->actionable_ = std::move(actionable);
+    setActionable(d->actionable_.get());
+}
+
 } // namespace fcitx
